#!/bin/bash
# This script is used to start the celery beat
# Automatically generated by acoupi, DO NOT EDIT THIS FILE
# Date: {{ now }}

{{ celery_bin }} \
    -A {{ app_name }} \
	beat \
	--pidfile={{ run_dir }}/beat.pid \
    --logfile={{ log_dir }}/beat.log \
	--loglevel="{{ log_level }}"
